#include	"trace.h"

/* defaults */
int	datalen = 32;	
int	max_ttl = 30;
int	nprobes = 3;

int debug;

u_short	dport = 32768 + 666;


void help_msg(char *name) {
  printf("usage: %s  [ -m <time to live> -q <queries> -d ] <hostname>\n",name);
  printf("\t -d turn on the debugging \n");
  return;
}

char * Sock_ntop_host(const struct sockaddr *sa, socklen_t salen)
{
  static char str[128];
  struct sockaddr_in	*sin = (struct sockaddr_in *) sa;
  
  if (inet_ntop(AF_INET, &sin->sin_addr, str, sizeof(str)) == NULL)
    return(NULL);
  
  return(str);
}

struct addrinfo * Host_serv(const char *host, const char *serv, int family, int socktype)
{
  int    n;
  struct addrinfo	hints, *res;
  
  bzero(&hints, sizeof(struct addrinfo));
  hints.ai_flags = AI_CANONNAME;	/* always return canonical name */
  hints.ai_family = family;		/* 0, AF_INET, AF_INET6, etc */
  hints.ai_socktype = socktype;		/* 0, SOCK_STREAM, SOCK_DGRAM, etc */

  if ( (n = getaddrinfo(host, serv, &hints, &res)) != 0)
  {
    printf("host_serv error for %s, %s: %s\n", (host == NULL) ? "(no hostname)" : host,
	   (serv == NULL) ? "(no service name)" : serv, gai_strerror(n));
    return NULL;
  }
  return(res);	/* return pointer to first on linked list */
}


char * icmpcode(int code)
{
  switch (code)
  {
  case  0:        return("network unreachable");
  case  1:        return("host unreachable");
  case  2:        return("protocol unreachable");
  case  3:        return("port unreachable");
  case  4:        return("fragmentation required but DF bit set");
  case  5:        return("source route failed");
  case  6:        return("destination network unknown");
  case  7:        return("destination host unknown");
  case  8:        return("source host isolated (obsolete)");
  case  9:        return("destination network administratively prohibited");
  case 10:        return("destination host administratively prohibited");
  case 11:        return("network unreachable for TOS");
  case 12:        return("host unreachable for TOS");
  case 13:        return("communication administratively prohibited by filtering");
  case 14:        return("host recedence violation");
  case 15:        return("precedence cutoff in effect");
  default:        return("[unknown code]");
  }
}


void sigAlarm(int sigtype)
{
  return;         /* just interrupt the recvfrom() */
}

/*
** Return: -4 error
**         -3 on timeout
**         -2 on ICMP time exceeded in transit (caller keeps going)
**         -1 on ICMP port unreachable (caller is done)
**         >= 0 return value is some other ICMP unreachable code
*/

int mrecv(int seq, struct timeval *tv)
{
  int hlen1, hlen2, icmplen;
  socklen_t len;
  ssize_t n;
  struct ip *ip, *hip;
  struct icmp *icmp;
  struct udphdr *udp;
  
  alarm(3);

  for ( ; ; ) {
    len = sizeof(struct sockaddr_in);
    n = recvfrom(recvfd, recvbuf, sizeof(recvbuf), 0,(struct sockaddr *) &myrecv, &len);
    if (n < 0) {
      if (errno == EINTR)
	return(-3);             /* alarm expired */
      else
        {
	  printf("recvfrom error");
          return (-4);
        }
    }

    gettimeofday(tv, (struct timezone *) 0);         /* get time of packet arrival */
    
    ip = (struct ip *) recvbuf;     /* start of IP header */
    hlen1 = ip->ip_hl << 2;         /* length of IP header */
    
    icmp = (struct icmp *) (recvbuf + hlen1); /* start of ICMP header */
    if ( (icmplen = n - hlen1) < 8)
       {
         printf("icmplen (%d) < 8", icmplen);
         return -4;
       }
    
    if (icmp->icmp_type == ICMP_TIMXCEED && icmp->icmp_code == ICMP_TIMXCEED_INTRANS) {
      if (icmplen < 8 + 20 + 8)
       {
	printf("icmplen (%d) < 8 + 20 + 8", icmplen);
        return -4;
       }

      hip = (struct ip *) (recvbuf + hlen1 + 8);
      hlen2 = hip->ip_hl << 2;
      udp = (struct udphdr *) (recvbuf + hlen1 + 8 + hlen2);
      if (hip->ip_p == IPPROTO_UDP && udp->source == htons(sport) && udp->dest == htons(dport + seq))
	return(-2);             /* we hit an intermediate router */
      
    } else if (icmp->icmp_type == ICMP_UNREACH) {
      if (icmplen < 8 + 20 + 8)
       {
	printf("icmplen (%d) < 8 + 20 + 8", icmplen);
        return -4;
       }
      
      hip = (struct ip *) (recvbuf + hlen1 + 8);
      hlen2 = hip->ip_hl << 2;
      udp = (struct udphdr *) (recvbuf + hlen1 + 8 + hlen2);
      if (hip->ip_p == IPPROTO_UDP && udp->source == htons(sport) && udp->dest == htons(dport + seq)) {
	if (icmp->icmp_code == ICMP_UNREACH_PORT)
	  return(-1);     /* have reached destination */
	else
	  return(icmp->icmp_code);        /* 0, 1, 2, ... */
      }
    }
  }
}


int sock_cmp_addr(const struct sockaddr_in *sa1, const struct sockaddr_in *sa2, socklen_t salen)
{
  if (sa1->sin_family != sa2->sin_family )
     return -1;

  if (sa1->sin_family == AF_INET)
    return(memcmp( &sa1->sin_addr, &sa2->sin_addr, sizeof(struct in_addr)));
  
  return (-1);
}

void tv_sub(struct timeval *out, struct timeval *in)
{
  if ( (out->tv_usec -= in->tv_usec) < 0) {       /* out -= in */
    --out->tv_sec;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

/*
** check if some data was send to us hopefully our icmp packet
*/
int waitforreply(int sockfd)
{
  int maxfd;
  int err;
 
  struct timeval  tv;
  fd_set   readfds;
 
  FD_ZERO(&readfds);
  maxfd=sockfd+1;
  FD_SET(sockfd,&readfds);
 
  tv.tv_sec=3;
  tv.tv_usec=0;
 
  err= select(maxfd,&readfds,NULL,NULL,&tv) ;
  if (err == 0)  return 0;
 
 if(FD_ISSET(sockfd,&readfds))
   return 1;

 return 0;
}

void traceloop(void)
{
  int     seq, code, done;
  double  rtt;
  struct rec      *rec;
  struct timeval  tvrecv;
  int sendfd;
  int ttl;
  int probe;
  
  recvfd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  setuid(getuid());               /* don't need special permissions any more */
  
  sendfd = socket(AF_INET, SOCK_DGRAM, 0);

  mybind.sin_family = AF_INET;
  sport = (pid & 0xffff) | 0x8000;   /* our source UDP port# */
  mybind.sin_port = htons(sport);
  bind(sendfd,(struct sockaddr *) &mybind, sizeof(mybind) );
  
  seq = 0;
  done = 0;
  for (ttl = 1; ttl <= max_ttl && done == 0; ttl++) {
     setsockopt(sendfd, IPPROTO_IP, IP_TTL , &ttl, sizeof(int));
     memset(&mylast,0,sizeof(mylast));
	  
     printf("%2d  ", ttl);
     fflush(stdout);
	  
     for (probe = 0; probe < nprobes; probe++) {
        alarm(0); /* cancel the timer */
        rec = (struct rec *) sendbuf;
        rec->rec_seq = ++seq;
        rec->rec_ttl = ttl;
        gettimeofday(&rec->rec_tv, (struct timezone *) 0);
	    
        mysend.sin_port = htons(dport+seq);

      for (code = sizeof(struct rec); code <datalen ; code++)
        sendbuf[code] = ' '+code;

    
       if( sendto(sendfd, sendbuf, datalen, 0, (struct sockaddr *) &mysend, sizeof(mysend) ) < 0)
       {
         perror("sendto ");
         return;
       }

      if ( waitforreply(recvfd) == 0)
         {
           (void) printf( " *");
           continue;
         }

      if( (code = mrecv(seq, &tvrecv)) == -3)
       {
         printf(" *");           /* timeout, no reply */
       }
	    else {
	      char    str[NI_MAXHOST];
              if(code == -4) return;
	      
	      if ( sock_cmp_addr(&myrecv,&mylast, sizeof(struct sockaddr_in)) != 0) {
		if (getnameinfo((struct sockaddr *) &myrecv,sizeof(struct sockaddr_in) , str, sizeof(str), NULL, 0, 0) == 0)
		  printf(" %s (%s)", str, Sock_ntop_host((struct sockaddr *) &myrecv, sizeof(struct sockaddr_in) ));
		else
		  printf(" %s",
			 Sock_ntop_host((struct sockaddr *) &myrecv, sizeof(struct sockaddr_in) ));

		memcpy(&mylast, &myrecv, sizeof(myrecv));
	      }

	      tv_sub(&tvrecv, &rec->rec_tv);
	      rtt = tvrecv.tv_sec * 1000.0 + tvrecv.tv_usec / 1000.0;
	      printf("  %.3f ms", rtt);
	      
	      if (code == -1)         /* port unreachable; at destination */
		done++;
	      else if (code >= 0)
		printf(" (ICMP %s)", icmpcode(code));
	    }
	    fflush(stdout);
	  }
	  printf("\n");
        }
}

int main(int argc, char **argv)
{
  int	c;
  struct addrinfo	*ai;
  char *host=NULL;

  opterr = 0;
  while ( (c = getopt(argc, argv, "m:q:d")) != -1) {
    switch (c) {
     case 'd':
          debug = 1;
          if (debug)
            printf("turn on debug mode.\n");
          break;

    case 'm':
      if ( (max_ttl = atoi(optarg)) <= 1)
       {
         help_msg(argv[0]);
         return 1;
       }

        if (debug)
            printf("time to live is specified to be %d.\n",max_ttl);

      break;

    case 'q':
      if ( (nprobes = atoi(optarg)) < 1)
       {
         help_msg(argv[0]);
         return 1;
       }

       if (debug)
           printf("no of probes is specified to be %d.\n",nprobes);

      break;

    default:
      printf("unrecognized option: %c", c);
        break;

    }
  }
  
  if (optind != argc -1)
   {
    help_msg(argv[0]);
    return 1;
   }

  host = argv[optind];
  
  pid = getpid();

  signal(SIGALRM,sigAlarm);
  
  ai = Host_serv(host, NULL, 0, 0);

  if (ai == NULL) return 1;

  printf("traceroute to %s (%s): %d hops max, %d data bytes\n",	 ai->ai_canonname, Sock_ntop_host(ai->ai_addr, ai->ai_addrlen),
	 max_ttl, datalen);
  
  memcpy(&mysend,(struct sockaddr_in *) ai->ai_addr, sizeof(struct sockaddr_in));		/* contains destination address */
  memcpy(&myrecv,(struct sockaddr_in *) ai->ai_addr, sizeof(struct sockaddr_in));		/* contains destination address */
  
  traceloop();

  return 1;
  
}
